#include "engine/sidechain/networkinputstreamworker.h"

NetworkInputStreamWorker::NetworkInputStreamWorker() {
}

void NetworkInputStreamWorker::setSourceFifo(FIFO<CSAMPLE>* pFifo) {
    (void)pFifo;
}
