IDI_ICON1               ICON    DISCARDABLE     "res/images/icons/ic_mixxx.ico"

#include <windows.h>

#include "mixxx.rc.include"

#define VER_PRODUCTNAME_STR         "Mixxx\0"
#define VER_FILEDESCRIPTION_STR     "Mixxx digital DJ software"
#define VER_COMPANYNAME_STR         "The Mixxx Development Team"
// \xA9 for (c) symbol. The Microsoft Resource Compiler compiler assumes Latin-1 strings
#define VER_LEGALCOPYRIGHT_STR      "\xA9 2001-" CUR_YEAR " Mixxx Development Team\0"
#define VER_ORIGINALFILENAME_STR    "Mixxx.exe"

#if defined(AMD64) || defined(INTEL64) || defined(EM64T) || defined(x86_64)
    #define BITS                    " x64"
#elif defined(IA64)
    #define BITS                    " Itanium"
#else
    #define BITS                    ""
#endif

#define VER_FILEVERSION_STR         MIXXX_VERSION BITS "\0"
#define VER_PRODUCTVERSION_STR      MIXXX_VERSION BITS "\0"

#ifndef DEBUG
    #define VER_DEBUG               0
#else
    #define VER_DEBUG               VS_FF_DEBUG
#endif

#ifndef PRERELEASE
    #define VER_PRERELEASE          0
#else
    #define VER_PRERELEASE          VS_FF_PRERELEASE
#endif

VS_VERSION_INFO VERSIONINFO
FILEVERSION    	VER_FILEVERSION
PRODUCTVERSION 	VER_PRODUCTVERSION
FILEFLAGSMASK  	VS_FFI_FILEFLAGSMASK
FILEFLAGS      	(VER_DEBUG|VER_PRERELEASE|VER_PATCHED)
FILEOS         	VOS__WINDOWS32
FILETYPE       	VFT_APP
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904E4"
        BEGIN
            VALUE "CompanyName",      VER_COMPANYNAME_STR
            VALUE "FileDescription",  VER_FILEDESCRIPTION_STR
            VALUE "FileVersion",      VER_FILEVERSION_STR
            //VALUE "InternalName",     VER_INTERNALNAME_STR
            VALUE "LegalCopyright",   VER_LEGALCOPYRIGHT_STR
            //VALUE "LegalTrademarks1", VER_LEGALTRADEMARKS1_STR
            //VALUE "LegalTrademarks2", VER_LEGALTRADEMARKS2_STR
            VALUE "OriginalFilename", VER_ORIGINALFILENAME_STR
            VALUE "ProductName",      VER_PRODUCTNAME_STR
            VALUE "ProductVersion",   VER_PRODUCTVERSION_STR
        END
    END

    BLOCK "VarFileInfo"
    BEGIN
        // Neutral Language (0x0) Unicode (1200).
        VALUE "Translation", 0x0, 1200

    END
END
