/**
* @file midienumerator.cpp
* @author Sean Pappalardo spappalardo@mixxx.org
* @date Tue 15 Mar 2012
* @brief Base class handling discovery and enumeration of DJ controllers that use the MIDI protocol.
*
*/

#include "controllers/midi/midienumerator.h"

#include "moc_midienumerator.cpp"

MidiEnumerator::MidiEnumerator() : ControllerEnumerator() {
}

MidiEnumerator::~MidiEnumerator() {
}
