#include "engineobject.h"

#include "moc_engineobject.cpp"

EngineObject::EngineObject() {
}

EngineObject::~EngineObject() {
}

EngineObjectConstIn::EngineObjectConstIn() {
}

EngineObjectConstIn::~EngineObjectConstIn() {
}
